@@ -1,15 +1,15 @@
 #Copyright ReportLab Europe Ltd. 2000-2017
 #see license.txt for license details
-#history http://www.reportlab.co.uk/cgi-bin/viewcvs.cgi/public/reportlab/trunk/reportlab/graphics/charts/textlabels.py
+#history https://bitbucket.org/rptlab/reportlab/history-node/tip/src/reportlab/graphics/charts/textlabels.py
 __version__='3.3.0'
 import string
 
 from reportlab.lib import colors
-from reportlab.lib.utils import simpleSplit, _simpleSplit
+from reportlab.lib.utils import simpleSplit, _simpleSplit, isBytes
 from reportlab.lib.validators import isNumber, isNumberOrNone, OneOf, isColorOrNone, isString, \
         isTextAnchor, isBoxAnchor, isBoolean, NoneOr, isInstanceOf, isNoneOrString, isNoneOrCallable
 from reportlab.lib.attrmap import *
-from reportlab.pdfbase.pdfmetrics import stringWidth, getAscentDescent
+from reportlab.pdfbase.pdfmetrics import stringWidth, getAscentDescent, getFont, unicode2T1
 from reportlab.graphics.shapes import Drawing, Group, Circle, Rect, String, STATE_DEFAULTS
 from reportlab.graphics.shapes import _PATH_OP_ARG_COUNT, _PATH_OP_NAMES, definePath
 from reportlab.graphics.widgetbase import Widget, PropHolder
@@ -61,17 +61,39 @@ def _processGlyph(G, truncate=1, pathReverse=0):
 def _text2PathDescription(text, x=0, y=0, fontName=_baseGFontName, fontSize=1000,
                             anchor='start', truncate=1, pathReverse=0):
     from reportlab.graphics import renderPM, _renderPM
-    _gs = _renderPM.gstate(1,1)
-    renderPM._setFont(_gs,fontName,fontSize)
+    font = getFont(fontName)
+    if font._multiByte and not font._dynamicFont:
+        raise ValueError("_text2PathDescription doesn't support multi byte fonts like %r" % fontName)
     P = []
     if not anchor=='start':
-        textLen = stringWidth(text, fontName,fontSize)
+        textLen = stringWidth(text, fontName, fontSize)
         if anchor=='end':
             x = x-textLen
         elif anchor=='middle':
             x = x - textLen/2.
-    for g in _gs._stringPath(text,x,y):
-        P.extend(_processGlyph(g,truncate=truncate,pathReverse=pathReverse))
+    _gs = _renderPM.gstate(1,1)
+    renderPM._setFont(_gs,fontName,fontSize)
+    if font._dynamicFont:
+        for g in _gs._stringPath(text,x,y):
+            P.extend(_processGlyph(g,truncate=truncate,pathReverse=pathReverse))
+    else:
+        if isBytes(text):
+            try:
+                text = text.decode('utf8')
+            except UnicodeDecodeError as e:
+                i,j = e.args[2:4]
+                raise UnicodeDecodeError(*(e.args[:4]+('%s\n%s-->%s<--%s' % (e.args[4],text[max(i-10,0):i],text[i:j],text[j:j+10]),)))
+        fc = font
+        FT = unicode2T1(text,[font]+font.substitutionFonts)
+        nm1 = len(FT)-1
+        for i, (f, t) in enumerate(FT):
+            if f!=fc:
+                renderPM._setFont(_gs,f.fontName,fontSize)
+                fc = f
+            for g in _gs._stringPath(t,x,y):
+                P.extend(_processGlyph(g,truncate=truncate,pathReverse=pathReverse))
+            if i!=nm1:
+                x += f.stringWidth(t.decode(f.encName), fontSize)
     return P
 
 def _text2Path(text, x=0, y=0, fontName=_baseGFontName, fontSize=1000,
@@ -200,51 +222,53 @@ class Label(Widget):
             ba = _A2BA[ba[-1]][na]
         return ba
 
-    def computeSize(self):
-        # the thing will draw in its own coordinate system
-        self._lineWidths = []
-        topPadding = self.topPadding
-        leftPadding = self.leftPadding
-        rightPadding = self.rightPadding
-        bottomPadding = self.bottomPadding
-        self._lines = simpleSplit(self._text,self.fontName,self.fontSize,self.maxWidth)
-        if not self.width:
-            self._width = leftPadding+rightPadding
-            if self._lines:
-                self._lineWidths = [stringWidth(line,self.fontName,self.fontSize) for line in self._lines]
-                self._width += max(self._lineWidths)
-        else:
-            self._width = self.width
+    def _getBaseLineRatio(self):
         if self.useAscentDescent:
             self._ascent, self._descent = getAscentDescent(self.fontName,self.fontSize)
             self._baselineRatio = self._ascent/(self._ascent-self._descent)
         else:
             self._baselineRatio = 1/1.2
+
+    def _computeSizeEnd(self,objH):
+        self._height = self.height or (objH + self.topPadding + self.bottomPadding)
+        self._ewidth = (self._width-self.leftPadding-self.rightPadding)
+        self._eheight = (self._height-self.topPadding-self.bottomPadding)
+        boxAnchor = self._getBoxAnchor()
+        if boxAnchor in ['n','ne','nw']:
+            self._top = -self.topPadding
+        elif boxAnchor in ['s','sw','se']:
+            self._top = self._height-self.topPadding
+        else:
+            self._top = 0.5*self._eheight
+        self._bottom = self._top - self._eheight
+
+        if boxAnchor in ['ne','e','se']:
+            self._left = self.leftPadding - self._width
+        elif boxAnchor in ['nw','w','sw']:
+            self._left = self.leftPadding
+        else:
+            self._left = -self._ewidth*0.5
+        self._right = self._left+self._ewidth
+
+    def computeSize(self):
+        # the thing will draw in its own coordinate system
+        self._lineWidths = []
+        self._lines = simpleSplit(self._text,self.fontName,self.fontSize,self.maxWidth)
+        if not self.width:
+            self._width = self.leftPadding+self.rightPadding
+            if self._lines:
+                self._lineWidths = [stringWidth(line,self.fontName,self.fontSize) for line in self._lines]
+                self._width += max(self._lineWidths)
+        else:
+            self._width = self.width
+        self._getBaseLineRatio()
         if self.leading:
             self._leading = self.leading
         elif self.useAscentDescent:
             self._leading = self._ascent - self._descent
         else:
             self._leading = self.fontSize*1.2
-        self._height = self.height or (self._leading*len(self._lines) + topPadding + bottomPadding)
-        self._ewidth = (self._width-leftPadding-rightPadding)
-        self._eheight = (self._height-topPadding-bottomPadding)
-        boxAnchor = self._getBoxAnchor()
-        if boxAnchor in ['n','ne','nw']:
-            self._top = -topPadding
-        elif boxAnchor in ['s','sw','se']:
-            self._top = self._height-topPadding
-        else:
-            self._top = 0.5*self._eheight
-        self._bottom = self._top - self._eheight
-
-        if boxAnchor in ['ne','e','se']:
-            self._left = leftPadding - self._width
-        elif boxAnchor in ['nw','w','sw']:
-            self._left = leftPadding
-        else:
-            self._left = -self._ewidth*0.5
-        self._right = self._left+self._ewidth
+        self._computeSizeEnd(self._leading*len(self._lines))
 
     def _getTextAnchor(self):
         '''This can be overridden to allow special effects'''
@@ -464,3 +488,86 @@ class RedNegativeChanger(CustomDrawChanger):
             R['fillColor'] = obj.fillColor
             obj.fillColor = self.fillColor
         return R
+
+from reportlab.platypus import XPreformatted, Paragraph
+from reportlab.lib.styles import ParagraphStyle
+from reportlab.lib.enums import TA_LEFT, TA_RIGHT, TA_CENTER
+_ta2al = dict(start=TA_LEFT,end=TA_RIGHT,middle=TA_RIGHT)
+try:
+    from rlextra.graphics.canvasadapter import DirectDrawFlowable
+except ImportError:
+    DirectDrawFlowable = None
+
+class XLabel(Label):
+    '''like label but uses XPreFormatted/Paragraph to draw the _text'''
+    _attrMap = AttrMap(BASE=Label,
+            )
+
+    def __init__(self,*args,**kwds):
+        self._flowableClass = kwds.pop('flowableClass',XPreformatted)
+        ddf = kwds.pop('directDrawClass',DirectDrawFlowable)
+        if ddf is None:
+            raise RuntimeError('DirectDrawFlowable class is not available you need the rlextra package as well as reportlab')
+        self._ddf = ddf
+        Label.__init__(self,*args,**kwds)
+
+    def computeSize(self):
+        # the thing will draw in its own coordinate system
+        self._lineWidths = []
+        sty = self._style = ParagraphStyle('xlabel-generated',
+                fontName=self.fontName,
+                fontSize=self.fontSize,
+                fillColor=self.fillColor,
+                strokeColor=self.strokeColor,
+                )
+        self._getBaseLineRatio()
+        if self.useAscentDescent:
+            sty.autoLeading = True
+            sty.leading = self._ascent - self._descent
+        else:
+            sty.leading = self.leading if self.leading else self.fontSize*1.2
+        self._leading = sty.leading
+        sty.alignment = _ta2al[self._getTextAnchor()]
+        self._obj = obj = self._flowableClass(self._text,style=sty)
+        _, objH = obj.wrap(self.maxWidth or 0x7fffffff,0x7fffffff)
+
+        if not self.width:
+            self._width = self.leftPadding+self.rightPadding
+            self._width += self._obj._width_max
+        else:
+            self._width = self.width
+        self._computeSizeEnd(objH)
+
+    def _rawDraw(self):
+        _text = self._text
+        self._text = _text or ''
+        self.computeSize()
+        self._text = _text
+        g = Group()
+        g.translate(self.x + self.dx, self.y + self.dy)
+        g.rotate(self.angle)
+
+        textAnchor = self._getTextAnchor()
+        if textAnchor == 'start':
+            x = self._left
+        elif textAnchor == 'middle':
+            x = self._left + self._ewidth*0.5
+        else:
+            x = self._right
+
+        # paint box behind text just in case they
+        # fill it
+        if self.boxFillColor or (self.boxStrokeColor and self.boxStrokeWidth):
+            g.add(Rect( self._left-self.leftPadding,
+                        self._bottom-self.bottomPadding,
+                        self._width,
+                        self._height,
+                        strokeColor=self.boxStrokeColor,
+                        strokeWidth=self.boxStrokeWidth,
+                        fillColor=self.boxFillColor)
+                        )
+        g1 = Group()
+        g1.translate(x,self._top-self._eheight)
+        g1.add(self._ddf(self._obj))
+        g.add(g1)
+        return g
