@@ -16,11 +16,12 @@ import hashlib
 from string import digits
 import tempfile
 from math import sin, cos, tan, pi, ceil
-from reportlab import rl_config
+from reportlab import rl_config, ascii, xrange
 from reportlab.pdfbase import pdfutils
 from reportlab.pdfbase import pdfdoc
 from reportlab.pdfbase import pdfmetrics
-from reportlab.pdfgen  import pdfgeom, pathobject, textobject
+from reportlab.pdfgen  import pdfgeom, pathobject
+from reportlab.pdfgen.textobject import PDFTextObject, _PDFColorSetter
 from reportlab.lib.colors import black, _chooseEnforceColorSpace, Color, CMYKColor, toColor
 from reportlab.lib.utils import import_zlib, ImageReader, isSeq, isStr, isUnicode, _digester
 from reportlab.lib.rl_accel import fp_str, escapePDF
@@ -110,7 +111,7 @@ def _buildColorFunction(colors, positions):
     # equally distribute if positions not specified
     if positions is None:
         nc = len(colors)
-        positions = [(1.0*x)/(nc-1) for x in range(nc)]
+        positions = [float(x)/(nc-1) for x in xrange(nc)]
     else:
         # sort positions and colors in increasing order
         poscolors = list(zip(positions, colors))
@@ -177,7 +178,7 @@ class   ExtGState:
         x._c = self._c
         return x
 
-class Canvas(textobject._PDFColorSetter):
+class Canvas(_PDFColorSetter):
     """This class is the programmer's interface to the PDF file format.  Methods
     are (or will be) provided here to do just about everything PDF can do.
 
@@ -390,7 +391,7 @@ class Canvas(textobject._PDFColorSetter):
      _charSpace _wordSpace _horizScale _textRenderMode _rise _textLineMatrix
      _textMatrix _lineCap _lineJoin _lineDash _lineWidth _mitreLimit _fillColorObj
      _strokeColorObj _extgstate""".split()
-    STATE_RANGE = list(range(len(STATE_ATTRIBUTES)))
+    STATE_RANGE = list(xrange(len(STATE_ATTRIBUTES)))
 
         #self._addStandardFonts()
 
@@ -519,6 +520,10 @@ class Canvas(textobject._PDFColorSetter):
            which is passing data into ReportLab, if you wish to name it."""
         self._doc.setCreator(creator)
 
+    def setProducer(self, producer):
+        """change the default producer value"""
+        self._doc.setProducer(producer)
+
     def setKeywords(self, keywords):
         """write a list of keywords into the PDF file which shows in document properties.
         Either submit a single string or a list/tuple"""
@@ -833,7 +838,7 @@ class Canvas(textobject._PDFColorSetter):
     #
     ######################################################
     def drawInlineImage(self, image, x,y, width=None,height=None,
-            preserveAspectRatio=False,anchor='c'):
+            preserveAspectRatio=False,anchor='c', anchorAtXY=False, showBoundary=False):
         """See drawImage, which should normally be used instead... 
         
         drawInlineImage behaves like drawImage, but stores the image content
@@ -852,11 +857,11 @@ class Canvas(textobject._PDFColorSetter):
         img_obj = PDFImage(image, x,y, width, height)
         img_obj.drawInlineImage(self,
             preserveAspectRatio=preserveAspectRatio, 
-            anchor=anchor)
+            anchor=anchor,anchorAtXY=anchorAtXY,showBoundary=showBoundary)
         return (img_obj.width, img_obj.height)
 
     def drawImage(self, image, x, y, width=None, height=None, mask=None, 
-            preserveAspectRatio=False, anchor='c'):
+            preserveAspectRatio=False, anchor='c', anchorAtXY=False, showBoundary=False):
         """Draws the image (ImageReader object or filename) as specified.
 
         "image" may be an image filename or an ImageReader object. 
@@ -960,7 +965,7 @@ class Canvas(textobject._PDFColorSetter):
                 del imgObj._smask
 
         # ensure we have a size, as PDF will make it 1x1 pixel otherwise!
-        x,y,width,height,scaled = aspectRatioFix(preserveAspectRatio,anchor,x,y,width,height,imgObj.width,imgObj.height)
+        x,y,width,height,scaled = aspectRatioFix(preserveAspectRatio,anchor,x,y,width,height,imgObj.width,imgObj.height,anchorAtXY)
 
         # scale and draw
         self.saveState()
@@ -968,6 +973,8 @@ class Canvas(textobject._PDFColorSetter):
         self.scale(width, height)
         self._code.append("/%s Do" % regName)
         self.restoreState()
+        if showBoundary:
+            self.rect(x,y,width,height,stroke=1,fill=0)
 
         # track what's been used on this page
         self._formsinuse.append(name)
@@ -977,9 +984,7 @@ class Canvas(textobject._PDFColorSetter):
     def _restartAccumulators(self):
         if self._codeStack:
             # restore the saved code
-            saved = self._codeStack[-1]
-            del self._codeStack[-1]
-            self._code, self._formsinuse, self._annotationrefs, self._formData,self._colorsUsed, self._shadingUsed = saved
+            self._code, self._formsinuse, self._annotationrefs, self._formData,self._colorsUsed, self._shadingUsed = self._codeStack.pop(-1)
         else:
             self._code = []    # ready for more...
             self._psCommandsAfterPage = []
@@ -1424,6 +1429,23 @@ class Canvas(textobject._PDFColorSetter):
             self._code.append('%s m %s l' % (fp_str(x1, y1), fp_str(x2, y2)))
         self._code.append('S')
 
+    def cross(self, x, y, size=5, gap=1, text=None, strokeColor=None, strokeWidth=None, fontSize=3):
+        size *= 0.5
+        gap *= 0.5
+        crosshairs = [(x,y-size,x,y-gap),(x,y+gap,x,y+size), (x-size,y,x-gap,y),(x+gap,y,x+size,y)]
+        self.saveState()
+        if strokeColor:
+            self.setStrokeColor(strokeColor)
+            if text:
+                self.setFillColor(strokeColor)
+        if strokeWidth:
+            self.setLineWidth(StrokeWidth)
+        self.lines(crosshairs)
+        if text:
+            if fontSize is not None: self.setFontSize(fontSize)
+            self.drawRightString(x-size, y, text)
+        self.restoreState()
+
     def grid(self, xlist, ylist):
         """Lays out a grid in current line style.  Supply list of
         x an y positions."""
@@ -1536,34 +1558,39 @@ class Canvas(textobject._PDFColorSetter):
         # use PDFTextObject for multi-line text.
         ##################################################
 
-    def drawString(self, x, y, text, mode=None, charSpace=0):
+    def drawString(self, x, y, text, mode=None, charSpace=0, direction=None, wordSpace=None):
         """Draws a string in the current text styles."""
         if sys.version_info[0] == 3 and not isinstance(text, str):
             text = text.decode('utf-8')
         #we could inline this for speed if needed
-        t = self.beginText(x, y)
+        t = self.beginText(x, y, direction=direction)
         if mode is not None: t.setTextRenderMode(mode)
         if charSpace: t.setCharSpace(charSpace)
+        if wordSpace: t.setWordSpace(wordSpace)
         t.textLine(text)
         if charSpace: t.setCharSpace(0)
+        if wordSpace: t.setWordSpace(0)
         if mode is not None: t.setTextRenderMode(0)
         self.drawText(t)
 
-    def drawRightString(self, x, y, text, mode=None, charSpace=0):
+    def drawRightString(self, x, y, text, mode=None, charSpace=0, direction=None, wordSpace=None):
         """Draws a string right-aligned with the x coordinate"""
         if sys.version_info[0] == 3 and not isinstance(text, str):
             text = text.decode('utf-8')
         width = self.stringWidth(text, self._fontname, self._fontsize)
         if charSpace: width += (len(text)-1)*charSpace
-        t = self.beginText(x - width, y)
+        if wordSpace: width += (text.count(u' ')+text.count(u'\xa0')-1)*wordSpace
+        t = self.beginText(x - width, y, direction=direction)
         if mode is not None: t.setTextRenderMode(mode)
         if charSpace: t.setCharSpace(charSpace)
+        if wordSpace: t.setWordSpace(wordSpace)
         t.textLine(text)
         if charSpace: t.setCharSpace(0)
+        if wordSpace: t.setWordSpace(0)
         if mode is not None: t.setTextRenderMode(0)
         self.drawText(t)
 
-    def drawCentredString(self, x, y, text, mode=None, charSpace=0):
+    def drawCentredString(self, x, y, text, mode=None, charSpace=0, direction=None, wordSpace=None):
         """Draws a string centred on the x coordinate. 
         
         We're British, dammit, and proud of our spelling!"""
@@ -1571,15 +1598,18 @@ class Canvas(textobject._PDFColorSetter):
             text = text.decode('utf-8')
         width = self.stringWidth(text, self._fontname, self._fontsize)
         if charSpace: width += (len(text)-1)*charSpace
-        t = self.beginText(x - 0.5*width, y)
+        if wordSpace: width += (text.count(u' ')+text.count(u'\xa0')-1)*wordSpace
+        t = self.beginText(x - 0.5*width, y, direction=direction)
         if mode is not None: t.setTextRenderMode(mode)
         if charSpace: t.setCharSpace(charSpace)
+        if wordSpace: t.setWordSpace(wordSpace)
         t.textLine(text)
         if charSpace: t.setCharSpace(0)
+        if wordSpace: t.setWordSpace(0)
         if mode is not None: t.setTextRenderMode(0)
         self.drawText(t)
 
-    def drawAlignedString(self, x, y, text, pivotChar=rl_config.decimalSymbol, mode=None, charSpace=0):
+    def drawAlignedString(self, x, y, text, pivotChar=rl_config.decimalSymbol, mode=None, charSpace=0, direction=None, wordSpace=None):
         """Draws a string aligned on the first '.' (or other pivot character).
 
         The centre position of the pivot character will be used as x.
@@ -1620,16 +1650,20 @@ class Canvas(textobject._PDFColorSetter):
                 rightText = leftText[-1] + rightText
                 leftText = leftText[0:-1]
 
-            self.drawRightString(x-0.5*pivW, y, leftText, mode=mode, charSpace=charSpace)
-            self.drawString(x-0.5*pivW, y, rightText, mode=mode, charSpace=charSpace)
+            self.drawRightString(x-0.5*pivW, y, leftText, mode=mode, charSpace=charSpace,
+                    direction=direction, wordSpace=wordSpace)
+            self.drawString(x-0.5*pivW, y, rightText, mode=mode, charSpace=charSpace,
+                    direction=direction, wordSpace=wordSpace)
 
         else:
             #normal case
             leftText = parts[0]
-            self.drawRightString(x-0.5*pivW, y, leftText, mode=mode, charSpace=charSpace)
+            self.drawRightString(x-0.5*pivW, y, leftText, mode=mode, charSpace=charSpace,
+                    direction=direction, wordSpace=wordSpace)
             if len(parts) > 1:
                 rightText = pivotChar + parts[1]
-                self.drawString(x-0.5*pivW, y, rightText, mode=mode, charSpace=charSpace)
+                self.drawString(x-0.5*pivW, y, rightText, mode=mode, charSpace=charSpace,
+                        direction=direction, wordSpace=wordSpace)
 
     def getAvailableFonts(self):
         """Returns the list of PostScript font names available.
@@ -1732,10 +1766,10 @@ class Canvas(textobject._PDFColorSetter):
         #                   + (self._fillMode == FILL_EVEN_ODD and ' W* ' or ' W ')
         #                   + PATH_OPS[stroke,fill,self._fillMode])
 
-    def beginText(self, x=0, y=0):
+    def beginText(self, x=0, y=0, direction=None):
         """Returns a fresh text object.  Text objects are used
-           to add large amounts of text.  See textobject.PDFTextObject"""
-        return textobject.PDFTextObject(self, x, y)
+           to add large amounts of text.  See PDFTextObject"""
+        return PDFTextObject(self, x, y, direction=direction)
 
     def drawText(self, aTextObject):
         """Draws a text object"""
@@ -1903,5 +1937,12 @@ class Canvas(textobject._PDFColorSetter):
             self._doc._catalog.AcroForm = self.AcroForm = AcroForm(self)
             return self.AcroForm
 
+    @property
+    def drawBoundary(self):
+        if not hasattr(self,'_drawBoundary'):
+            from reportlab.platypus import Frame
+            self._drawBoundary = lambda sb,x,y,w,h: Frame._drawBoundary(self,sb,x,y,w,h)
+        return self._drawBoundary
+
 if __name__ == '__main__':
     print('For test scripts, look in tests')
